@@ -34,7 +34,7 @@
 #include <mpir.h>
 #endif
 
-#include "features.h"
+#include "z_features.h"
 #include "zarith.h"
 
 #ifdef __cplusplus
@@ -187,6 +187,22 @@ extern "C" {
 #define Z_HI_INT     0x40000000
 #endif
 
+/* safe bounds for the length of a base n string fitting in a native
+   int. Defined as the result of (n - 2) log_base(2) with n = 64 or
+   32. 
+*/
+#ifdef ARCH_SIXTYFOUR
+#define Z_BASE16_LENGTH_OP 15
+#define Z_BASE10_LENGTH_OP 18
+#define Z_BASE8_LENGTH_OP 20
+#define Z_BASE2_LENGTH_OP 62
+#else
+#define Z_BASE16_LENGTH_OP 7
+#define Z_BASE10_LENGTH_OP 9
+#define Z_BASE8_LENGTH_OP 10
+#define Z_BASE2_LENGTH_OP 30
+#endif
+
 #define Z_LIMB_BITS  (8 * sizeof(mp_limb_t))
 
 
@@ -212,6 +228,7 @@ extern struct custom_operations ml_z_custom_ops;
 
 static double ml_z_2p32; /* 2 ^ 32 in double */
 
+#if  Z_PERFORM_CHECK
 /* for debugging: dump a mp_limb_t array */
 static void ml_z_dump(const char* msg, mp_limb_t* p, mp_size_t sz)
 {
@@ -226,7 +243,9 @@ static void ml_z_dump(const char* msg, mp_limb_t* p, mp_size_t sz)
   printf("\n");
   fflush(stdout);
 }
+#endif
 
+#if Z_PERFORM_CHECK
 /* for debugging: check invariant */
 void ml_z_check(const char* fn, int line, const char* arg, value v)
 {
@@ -275,7 +294,8 @@ void ml_z_check(const char* fn, int line, const char* arg, value v)
     exit(1);
   }
 }
-
+#endif
+  
 /* for debugging */
 #if Z_PERFORM_CHECK
 #define Z_CHECK(v) ml_z_check(__FUNCTION__, __LINE__, #v, v)
@@ -326,6 +346,13 @@ static inline mp_limb_t* ml_z_dup_limb(mp_limb_t* src, mp_size_t sz)
   return r;
 }
 
+
+#ifdef _MSC_VER
+#define MAYBE_UNUSED
+#else
+#define MAYBE_UNUSED (void)
+#endif
+
 /* given a z object, define:
    - ptr_arg: a pointer to the first mp_limb_t
    - size_arg: the number of mp-limb_t
@@ -336,10 +363,10 @@ static inline mp_limb_t* ml_z_dup_limb(mp_limb_t* src, mp_size_t sz)
   mp_limb_t loc_##arg, *ptr_##arg;                                      \
   mp_size_t size_##arg;                                                 \
   intnat sign_##arg;                                                    \
-  (void)loc_##arg;                                                      \
-  (void)ptr_##arg;                                                      \
-  (void)size_##arg;                                                     \
-  (void)sign_##arg;
+  MAYBE_UNUSED loc_##arg;                                                      \
+  MAYBE_UNUSED ptr_##arg;                                                      \
+  MAYBE_UNUSED size_##arg;                                                     \
+  MAYBE_UNUSED sign_##arg;
 
 #define Z_ARG(arg)                                                      \
   if (Is_long(arg)) {                                                   \
@@ -540,14 +567,29 @@ CAMLprim value ml_z_of_float(value v)
   return r;
 }
 
-CAMLprim value ml_z_of_string_base(value b, value v)
+CAMLprim value ml_z_of_substring_base(value b, value v, value offset, value length)
 {
   CAMLparam1(v);
   CAMLlocal1(r);
-  char *d = String_val(v);
+  intnat ofs = Long_val(offset);
+  intnat len = Long_val(length);
+  /* make sure the ofs/length make sense */
+  if (ofs < 0
+      || len < 0
+      || (intnat)caml_string_length(v) < ofs + len)
+    caml_invalid_argument("Z.of_substring_base: invalid offset or length");
+  /* process the string */
+  char *d = String_val(v) + ofs;
+  char *end = d + len;  
   mp_size_t i, sz, sz2;
   mp_limb_t sign = 0;
   intnat base = Long_val(b);
+  /* We allow [d] to advance beyond [end] while parsing the prefix: 
+     sign, base, and/or leading zeros.  
+     This simplifies the code, and reading these locations is safe since
+     we don't progress beyond a terminating null character. 
+     At the end of the prefix, if we ran past the end, we return 0. 
+  */
   /* get optional sign */
   if (*d == '-') { sign ^= Z_SIGN_MASK; d++; }
   if (*d == '+') d++;
@@ -562,21 +604,48 @@ CAMLprim value ml_z_of_string_base(value b, value v)
     }
   }
   if (base < 2 || base > 16) 
-    caml_invalid_argument("Z.of_string_base: base must be between 2 and 16");
+    caml_invalid_argument("Z.of_substring_base: base must be between 2 and 16");
   while (*d == '0') d++;
-  sz = strlen(d);
-  if (!sz) r = Val_long(0); 
-  else {
-    /* converts to sequence of digits */
-    char* dd = (char*)malloc(strlen(d)+1);
-    strcpy(dd,d);
+  /* sz is the length of the substring that has not been consumed above. */
+  sz = end - d;
+#if Z_USE_NATINT
+  if (sz <= 0) {
+    /* "+", "-", "0x" are parsed as 0. */
+    r = Val_long(0);
+  }
+  /* Process common case (fits into a native integer) */
+  else if ((base == 10 && sz <= Z_BASE10_LENGTH_OP)
+        || (base == 16 && sz <= Z_BASE16_LENGTH_OP)
+        || (base == 8  && sz <= Z_BASE8_LENGTH_OP)
+        || (base == 2  && sz <= Z_BASE2_LENGTH_OP)) {
+      Z_MARK_OP;
+      intnat ret = 0;
+      for (i = 0; i < sz; i++) {
+        int digit = 0;
+        if (d[i] >= '0' && d[i] <= '9') digit = d[i] - '0';
+        else if (d[i] >= 'a' && d[i] <= 'f') digit = d[i] - 'a' + 10;
+        else if (d[i] >= 'A' && d[i] <= 'F') digit = d[i] - 'A' + 10;
+        else caml_invalid_argument("Z.of_substring_base: invalid digit");
+        if (digit >= base) 
+          caml_invalid_argument("Z.of_substring_base: invalid digit");
+        ret = ret * base + digit;
+      }
+      r = Val_long(ret * (sign ? -1 : 1));
+  } else
+#endif
+  {
+     /* converts to sequence of digits */
+    char* dd = (char*)malloc(sz+1);
+    strncpy(dd,d,sz);
+    /* make sure that dd is nul terminated */
+    dd[sz] = 0;
     for (i = 0; i < sz; i++) {
       if (dd[i] >= '0' && dd[i] <= '9') dd[i] -= '0';
       else if (dd[i] >= 'a' && dd[i] <= 'f') dd[i] -= 'a' - 10;
       else if (dd[i] >= 'A' && dd[i] <= 'F') dd[i] -= 'A' - 10;
-      else caml_invalid_argument("Z.of_string_base: invalid number");
+      else caml_invalid_argument("Z.of_substring_base: invalid digit");
       if (dd[i] >= base) 
-        caml_invalid_argument("Z.of_string_base: invalid number");
+        caml_invalid_argument("Z.of_substring_base: invalid digit");
     }
     r = ml_z_alloc(1 + sz / (2 * sizeof(mp_limb_t)));
     sz2 = mpn_set_str(Z_LIMB(r), (unsigned char*)dd, sz, base);
@@ -693,36 +762,6 @@ CAMLprim value ml_z_to_int64(value v)
   return caml_copy_int64(x);
 }
 
-CAMLprim value ml_z_to_float(value v)
-{
-  double x, m;
-  mp_size_t i;
-  Z_DECL(v);
-  Z_MARK_OP;
-  Z_CHECK(v);
-  if (Is_long(v)) return caml_copy_double(Long_val(v));
-  Z_MARK_SLOW;
-  Z_ARG(v);
-  m = sign_v ? -1. : 1.;
-  x = 0.;
-  for (i = 0; i < size_v; i++) {
-#ifdef ARCH_SIXTYFOUR
-    /* split into two 32-bit numbers, as 64-bit integers may not fit 
-       exactly in a double
-       the cast to long is a work-around for gcc's bug 37544
-    */
-    x += m * (intnat)(ptr_v[i] & 0xffffffff);
-    m *= ml_z_2p32;
-    x += m * (intnat)((ptr_v[i] >> 32) & 0xffffffff);
-    m *= ml_z_2p32;
-#else
-    x += m * ptr_v[i];
-    m *= ml_z_2p32;
-#endif
-  }
-  return caml_copy_double(x);
-}
-
 /* XXX: characters that do not belong to the format are ignored, this departs
    from the classic printf behavior (it copies them in the output)
  */
@@ -838,6 +877,7 @@ CAMLprim value ml_z_extract(value arg, value off, value len)
   value r;
   Z_DECL(arg);
   Z_MARK_OP;
+  MAYBE_UNUSED x;
   o = Long_val(off);
   l = Long_val(len);
   if (o < 0) caml_invalid_argument("Z.extract: negative bit offset");
@@ -1701,11 +1741,12 @@ CAMLprim value ml_z_gcd(value arg1, value arg2)
     if (!a1 || !a2) ml_z_raise_divide_by_zero();
     if (a1 < 0) a1 = -a1;
     if (a2 < 0) a2 = -a2;
+    if (a1 < a2) { intnat t = a1; a1 = a2; a2 = t; }
     while (a2)  {
       intnat r = a1 % a2;
       a1 = a2; a2 = r;
     }
-    if (Z_FITS_INT(a1)) return Val_long(a1);
+    return Val_long(a1);
   }
 #endif
   /* mpn_ version */
@@ -2268,6 +2309,125 @@ CAMLprim value ml_z_shift_right_trunc(value arg, value count)
   }
 }
 
+/* Helper function for numbits: number of leading 0 bits in x */
+
+#ifdef _LONG_LONG_LIMB
+#define BUILTIN_CLZ __builtin_clzll
+#else
+#define BUILTIN_CLZ __builtin_clzl
+#endif
+
+/* Use GCC or Clang built-in if available.  The argument must be != 0. */
+#if defined(__clang__) || __GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
+#define ml_z_clz BUILTIN_CLZ
+#else
+/* Portable C implementation - Hacker's Delight fig 5.12 */
+int ml_z_clz(mp_limb_t x)
+{
+  int n;
+  mp_limb_t y;
+#ifdef ARCH_SIXTYFOUR
+  n = 64;
+  y = x >> 32;  if (y != 0) { n = n - 32; x = y; }
+#else
+  n = 32;
+#endif
+  y = x >> 16;  if (y != 0) { n = n - 16; x = y; }
+  y = x >>  8;  if (y != 0) { n = n -  8; x = y; }
+  y = x >>  4;  if (y != 0) { n = n -  4; x = y; }
+  y = x >>  2;  if (y != 0) { n = n -  2; x = y; }
+  y = x >>  1;  if (y != 0) return n - 2;
+  return n - x;
+}
+#endif
+
+CAMLprim value ml_z_numbits(value arg)
+{
+  Z_DECL(arg);
+  intnat r;
+  int n;
+  Z_MARK_OP;
+  Z_CHECK(arg);
+#if Z_FAST_PATH
+  if (Is_long(arg)) {
+    /* fast path */
+    r = Long_val(arg);
+    if (r == 0) {
+      return Val_int(0);
+    } else {
+      n = ml_z_clz(r > 0 ? r : -r);
+      return Val_long(sizeof(intnat) * 8 - n);
+    }
+  }
+#endif
+  /* mpn_ version */  
+  Z_MARK_SLOW;
+  Z_ARG(arg);
+  if (size_arg == 0) return Val_int(0);
+  n = ml_z_clz(ptr_arg[size_arg - 1]);
+  return Val_long(size_arg * Z_LIMB_BITS - n);
+}
+
+/* Helper function for trailing_zeros: number of trailing 0 bits in x */
+
+#ifdef _LONG_LONG_LIMB
+#define BUILTIN_CTZ __builtin_ctzll
+#else
+#define BUILTIN_CTZ __builtin_ctzl
+#endif
+
+/* Use GCC or Clang built-in if available.  The argument must be != 0. */
+#if defined(__clang__) || __GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
+#define ml_z_ctz BUILTIN_CTZ
+#else
+/* Portable C implementation - Hacker's Delight fig 5.21 */
+int ml_z_ctz(mp_limb_t x)
+{
+  int n;
+  mp_limb_t y;
+  CAMLassert (x != 0);
+#ifdef ARCH_SIXTYFOUR
+  n = 63;
+  y = x << 32;  if (y != 0) { n = n - 32; x = y; }
+#else
+  n = 31;
+#endif
+  y = x << 16;  if (y != 0) { n = n - 16; x = y; }
+  y = x <<  8;  if (y != 0) { n = n -  8; x = y; }
+  y = x <<  4;  if (y != 0) { n = n -  4; x = y; }
+  y = x <<  2;  if (y != 0) { n = n -  2; x = y; }
+  y = x <<  1;  if (y != 0) { n = n - 1; }
+  return n;
+}
+#endif
+
+CAMLprim value ml_z_trailing_zeros(value arg)
+{
+  Z_DECL(arg);
+  intnat r;
+  mp_size_t i;
+  Z_MARK_OP;
+  Z_CHECK(arg);
+#if Z_FAST_PATH
+  if (Is_long(arg)) {
+    /* fast path */
+    r = Long_val(arg);
+    if (r == 0) {
+      return Val_long (Max_long);
+    } else {
+      /* No need to take absolute value of r, as ctz(-x) = ctz(x) */
+      return Val_long (ml_z_ctz(r));
+    }
+  }
+#endif
+  /* mpn_ version */  
+  Z_MARK_SLOW;
+  Z_ARG(arg);
+  if (size_arg == 0) return Val_long (Max_long);
+  for (i = 0; ptr_arg[i] == 0; i++) /* skip */;
+  return Val_long(i * Z_LIMB_BITS + ml_z_ctz(ptr_arg[i]));
+}
+
 /* helper function for popcount & hamdist: number of bits at 1 in x */
 /* maybe we should use the mpn_ function even for small arguments, in case
    the CPU has a fast popcount opcode?
@@ -2358,6 +2518,43 @@ CAMLprim value ml_z_hamdist(value arg1, value arg2)
   return Val_long(r);
 }
 
+CAMLprim value ml_z_testbit(value arg, value index)
+{
+  Z_DECL(arg);
+  uintnat b_idx;
+  mp_size_t l_idx, i;
+  mp_limb_t limb;
+  Z_MARK_OP;
+  Z_CHECK(arg);
+  b_idx = Long_val(index); /* Caml code checked index >= 0 */
+#if Z_FAST_PATH
+  if (Is_long(arg)) {
+    if (b_idx >= Z_LIMB_BITS) b_idx = Z_LIMB_BITS - 1;
+    return Val_int((Long_val(arg) >> b_idx) & 1);
+  }
+#endif
+  Z_MARK_SLOW;
+  Z_ARG(arg);
+  l_idx = b_idx / Z_LIMB_BITS;
+  if (l_idx >= size_arg) return Val_bool(sign_arg);
+  limb = ptr_arg[l_idx];
+  if (sign_arg != 0) {
+    /* If arg is negative, its 2-complement representation is
+       bitnot(abs(arg) - 1).
+       If any of the limbs of abs(arg) below l_idx is nonzero, 
+       the carry from the decrement dies before reaching l_idx,
+       and we just test bitnot(limb).
+       If all the limbs below l_idx are zero, the carry from the
+       decrement propagates to l_idx,
+       and we test bitnot(limb - 1) = - limb. */
+    for (i = 0; i < l_idx; i++) {
+      if (ptr_arg[i] != 0) { limb = ~limb; goto extract; }
+    }
+    limb = -limb;
+  }
+ extract:
+  return Val_int((limb >> (b_idx % Z_LIMB_BITS)) & 1);
+}
 
 /*---------------------------------------------------
   FUNCTIONS BASED ON mpz_t
@@ -2391,8 +2588,14 @@ value ml_z_from_mpz(mpz_t op)
   return ml_z_reduce(r, sz, (mpz_sgn(op) >= 0) ? 0 : Z_SIGN_MASK);
 }
 
+#if __GNU_MP_VERSION >= 5
+/* not exported by gmp.h */
+extern void __gmpn_divexact (mp_ptr, mp_srcptr, mp_size_t, mp_srcptr, mp_size_t);
+#endif
+
 CAMLprim value ml_z_divexact(value arg1, value arg2)
 {
+  Z_DECL(arg1); Z_DECL(arg2);
   Z_MARK_OP;
   Z_CHECK(arg1); Z_CHECK(arg2);
 #if Z_FAST_PATH
@@ -2406,9 +2609,31 @@ CAMLprim value ml_z_divexact(value arg1, value arg2)
     if (Z_FITS_INT(q)) return Val_long(q);
   }
 #endif
-  /* mpz_ version */
   Z_MARK_SLOW;
+#if __GNU_MP_VERSION >= 5
   {
+    /* mpn_ version */
+    Z_ARG(arg1);
+    Z_ARG(arg2);
+    if (!size_arg2)
+      ml_z_raise_divide_by_zero();
+    if (size_arg1 < size_arg2)
+      return Val_long(0);
+    {
+      CAMLparam2(arg1,arg2);
+      CAMLlocal1(q);
+      q = ml_z_alloc(size_arg1 - size_arg2 + 1);
+      Z_REFRESH(arg1); Z_REFRESH(arg2);
+      __gmpn_divexact(Z_LIMB(q),
+                      ptr_arg1,  size_arg1, ptr_arg2, size_arg2);
+      q = ml_z_reduce(q, size_arg1 - size_arg2 + 1, sign_arg1 ^ sign_arg2);
+      Z_CHECK(q);
+      CAMLreturn(q);
+    }
+  }
+#else
+  {
+    /* mpz_ version */
     CAMLparam2(arg1,arg2);
     CAMLlocal1(r);
     mpz_t a,b;
@@ -2422,6 +2647,7 @@ CAMLprim value ml_z_divexact(value arg1, value arg2)
     mpz_clear(b);
     CAMLreturn(r);
   }
+#endif
 }
  
 CAMLprim value ml_z_powm(value base, value exp, value mod)
@@ -2448,6 +2674,40 @@ CAMLprim value ml_z_powm(value base, value exp, value mod)
   CAMLreturn(r);
 }
 
+CAMLprim value ml_z_powm_sec(value base, value exp, value mod)
+{
+#ifndef HAS_MPIR
+#if __GNU_MP_VERSION >= 5
+  CAMLparam3(base,exp,mod);
+  CAMLlocal1(r);
+  mpz_t mbase, mexp, mmod;
+  ml_z_mpz_init_set_z(mbase, base);
+  ml_z_mpz_init_set_z(mexp, exp);
+  ml_z_mpz_init_set_z(mmod, mod);
+  if (mpz_sgn(mexp) <= 0)
+    caml_invalid_argument("Z.powm_sec: exponent must be positive");
+  if (! mpz_odd_p(mmod))
+    caml_invalid_argument("Z.powm_sec: modulus must be odd");
+  mpz_powm_sec(mbase, mbase, mexp, mmod);
+  r = ml_z_from_mpz(mbase);
+  mpz_clear(mbase);
+  mpz_clear(mexp);
+  mpz_clear(mmod);
+  CAMLreturn(r);
+#else
+  MAYBE_UNUSED(base);
+  MAYBE_UNUSED(exp);
+  MAYBE_UNUSED(mod);
+  caml_invalid_argument("Z.powm_sec: not available, needs GMP version >= 5");
+#endif
+#else
+  MAYBE_UNUSED(base);
+  MAYBE_UNUSED(exp);
+  MAYBE_UNUSED(mod);
+  caml_invalid_argument("Z.powm_sec: not available in MPIR, needs GMP version >= 5");
+#endif
+}
+
 CAMLprim value ml_z_pow(value base, value exp)
 {
   CAMLparam2(base,exp);
